/* Secure FILL - by van Hauser / [THC], vh@campus.de
 *
 * Secure FILL overwrites all available free diskspace by creating a file,
 * wiping all free diskspace it gets and finally deleting the file.
 * For security reasons full 16kb blocks are written so that the whole block
 * on which the file(s) live are overwritten.
 * The option -s does 2 overwritings.
 * The option -s twice does a real security wipe for 39 times, flushing
 * the caches after every write. The wipe technique was propsed by Peter
 * Gutmann at Usenix '96 and includes 10 random overwrites plus 29 special
 * defined characters. Take a look at the paper of him, it's really worth
 * your time.
 *
 * Advice : use always the options -ssv
 *
 * Read the manual for limitations.
 *
 */

#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <dirent.h>

#define DIR_SEPERATOR '/'
#define FLUSH sync();

char *VERSION = "v1.3 by van Hauser / THC";
int verbose = 0;


int smash_it(char *filename, int mode) 
{
    unsigned char write_modes[27][3] = {
        {"\x55\x55\x55"}, {"\xaa\xaa\xaa"}, {"\x92\x49\x24"}, {"\x49\x24\x92"},
        {"\x24\x92\x49"}, {"\x00\x00\x00"}, {"\x11\x11\x11"}, {"\x22\x22\x22"},
        {"\x33\x33\x33"}, {"\x44\x44\x44"}, {"\x55\x55\x55"}, {"\x66\x66\x66"},
        {"\x77\x77\x77"}, {"\x88\x88\x88"}, {"\x99\x99\x99"}, {"\xaa\xaa\xaa"},
        {"\xbb\xbb\xbb"}, {"\xcc\xcc\xcc"}, {"\xdd\xdd\xdd"}, {"\xee\xee\xee"},
        {"\xff\xff\xff"}, {"\x92\x49\x24"}, {"\x49\x24\x92"}, {"\x24\x92\x49"},
        {"\x6d\xb6\xdb"}, {"\xb6\xdb\x6d"}, {"\xdb\x6d\xb6"}
    };

    unsigned char std_array[3] = "\xff\xff\xff";
    FILE *f;
    int turn;
    unsigned char ch;
        
/* open the file, get the filesize, calculate the numbers of needed writings */
    if (verbose) printf("Creating %s ", filename);
    f = fopen(filename, "w+b");

/* do the overwriting stuff */
    if (mode != 0) { 
        for (turn=0; turn<=35; turn++) {
            rewind(f);
            if ((mode == 1) && (turn > 0)) break;
            if ((turn>=5) && (turn<=31)) 
               do {} while (fwrite(&write_modes[turn-5][0], 1, 3, f) == 3);
            else {
               do {
                   ch = (unsigned char) (256.0*rand()/(RAND_MAX+1.0));
               } while (fwrite(&ch, 1, 1, f) == 1);
            }
            FLUSH   
            if (verbose) {
               printf("*");
               fflush(stdout);
            }
        }
        rewind(f); 
    }

    do {} while (fwrite(std_array, 1, 3, f) == 3);
    if (verbose) printf("*");

/* Soft Flush -> Write buffered data */
    fflush(f);
    (void) fclose(f);
/* Hard Flush -> Force cached data to be written to disk - the defines above! */
    FLUSH
   
    return 0;
}


void main (int argc, char *argv[])
{
    char newname[1024];
    int result;
    int secure = 0;
    int schleife;
    int turn;
    int counter;
    struct stat filestat;
        
    if (argc==1){
        printf("Syntax: %s [-ssv] directory\n", argv[0]);
        printf("%s does a secure overwrite of the free diskspace area in the directory.\n", argv[0]);
        printf("Option -ss does real secure overwrite (39 times). -v is verbose mode.\n");
        exit(1);
    }
    while (1) {
        result = getopt(argc, argv, "sSvVdD");
        if (result<0) break;
        switch (result) {
	    case 's' :
            case 'S' : secure++;
                       break;
	    case 'v' :
            case 'V' : verbose++;
                       break;
            case 'd' : 
            case 'D' : break;
            default :  exit(1);        
        }
    }
    schleife = optind;
    if (schleife >= argc) {
        printf("Syntax: %s [-sv] directory\n", argv[0]);
        printf("You need to specify the directory to write the tempfile to.\n");
        exit(1);
    }

    do {
       strncpy(newname, argv[schleife], sizeof(newname));    
       if (opendir(newname) == NULL) {
           printf("%s : not a directory\n",newname);
       } else {

/* Generate random unique name for tempfile */
	    srand(getpid()+getuid());

            if (newname[strlen(newname)-1] != DIR_SEPERATOR) {
               result = strlen(newname);
               newname[result] = DIR_SEPERATOR;
               newname[result+1] = 0;
	    }    
            (void)strcat(newname, "oooooooo.ooo");

            turn = 0; result = 0;
            while ((result >= 0)&&(turn<=250)) {
                for (counter=strlen(newname)-1; (newname[counter] != DIR_SEPERATOR); counter--) 
                    if (newname[counter] != '.') 
                       newname[counter] = 97+(int) (27.0*rand()/(RAND_MAX+1.0));
		result = stat(newname, &filestat);
		turn++;
            };
            if (result>=0) printf("%s : couldn't find a free filename\n",argv[schleife]);
            else {
               printf("Wiping now.\nIf you abort, be sure to remove the file %s\n",newname);
               fflush(stdout);
               result = smash_it(newname, secure);
               switch (result) {
                   case 0 : if (verbose) printf(" Removed\n");
                            break;
                   case 1 : printf("%s : no write permissions\n", argv[schleife]);
                            break;
                   default: printf("Unknown error\n");
               }
               (void) unlink(newname);
            }
       }

       schleife++;
    } while (schleife<argc);
    fflush(stdout);
}
