/* Secure RM - by van Hauser / [THC], vh@campus.de
 *
 * Secure ReMove first overwrites then renames and finally deletes the target
 * file(s) specified via parameters.
 * For security reasons full 32kb blocks are written so that the whole block
 * on which the file(s) live are overwritten. (change #define #BLOCKSIZE)
 * The option -s overwrites 2 times the data.
 * The option -s twice does a real security wipe for 39 times, flushing
 * the caches after every write. The wipe technique was propsed by Peter
 * Gutmann at Usenix '96 and includes 10 random overwrites plus 29 special
 * defined characters. Take a look at the paper of him, it's really worth
 * your time. 
 *
 * Advice : set "alias rm 'srm -ssv'"
 *
 * Read the manual for limitations.
 *
 */

#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>

#define BLOCKSIZE 32768
#define DIR_SEPERATOR '/'
#define FLUSH sync();

char *VERSION = "v1.3 by van Hauser / THC";
int verbose = 0;

int smash_it(char *filename, int mode) 
{
    unsigned char write_modes[27][3] = {
        {"\x55\x55\x55"}, {"\xaa\xaa\xaa"}, {"\x92\x49\x24"}, {"\x49\x24\x92"},
        {"\x24\x92\x49"}, {"\x00\x00\x00"}, {"\x11\x11\x11"}, {"\x22\x22\x22"},
        {"\x33\x33\x33"}, {"\x44\x44\x44"}, {"\x55\x55\x55"}, {"\x66\x66\x66"},
        {"\x77\x77\x77"}, {"\x88\x88\x88"}, {"\x99\x99\x99"}, {"\xaa\xaa\xaa"},
        {"\xbb\xbb\xbb"}, {"\xcc\xcc\xcc"}, {"\xdd\xdd\xdd"}, {"\xee\xee\xee"},
        {"\xff\xff\xff"}, {"\x92\x49\x24"}, {"\x49\x24\x92"}, {"\x24\x92\x49"},
        {"\x6d\xb6\xdb"}, {"\xb6\xdb\x6d"}, {"\xdb\x6d\xb6"}
    };

    unsigned char std_array[3] = "\xff\xff\xff";
    FILE *f;
    unsigned long writes;
    unsigned long counter;
    unsigned long filesize;
    struct stat filestat;
    int turn;
    int result;
    unsigned char ch;
    char newname[512];
        
/* open the file, get the filesize, calculate the numbers of needed writings */
    f = fopen(filename, "r+b");
    if (f == NULL) return 1;
    (void) stat(filename, &filestat);
    filesize = filestat.st_size;
    writes = (1 + (filesize / BLOCKSIZE)) * ((BLOCKSIZE+2)/3);

/* do the overwriting stuff */
    if (mode != 0) { 
        for (turn=0; turn<=35; turn++) {
            rewind(f); 
            if ((mode ==1) && (turn > 0)) break;
            if ((turn>=5) && (turn<=31)) {
               for (counter=1; counter<=writes; counter++) 
            if (fwrite(&write_modes[turn-5][0], 1, 3, f) != 3) {
               (void) fclose(f);
               return 2;
            }
	    } else {
	       for (counter=1; counter<=writes*3; counter++) {
 		  ch = (unsigned char) (256.0*rand()/(RAND_MAX+1.0));
	          fwrite(&ch, 1, 1, f); 
	       }  
	    }
            FLUSH   
            if (verbose) {
               printf("*");
               fflush(stdout);
            }
	}
	rewind(f);
    }

    for (counter=1; counter<=writes; counter++)
        if (fwrite(std_array, 1, 3, f) != 3) {
           (void) fclose(f);
           return 2;
        }
    if (verbose) {
       printf("*");
       fflush(stdout);
    }
                                                      
/* Soft Flush -> Write buffered data */
    fflush(f);
    (void) fclose(f);
/* Hard Flush -> Force cached data to be written to disk */
    FLUSH
   
/* Generate random unique name, renaming and deleting of the file */
    strncpy(newname, filename, sizeof(newname));
    turn = 0;
    do {
        for (counter=strlen(newname)-1; (newname[counter] != DIR_SEPERATOR)&&(counter>=0); counter--)
             if (newname[counter] != '.') 
                 newname[counter] = 97+(int) (27.0*rand()/(RAND_MAX+1.0));
        result = stat(newname, &filestat);
        turn++;
    } while ((result >= 0)&&(turn<=100));
    if (turn<=100) {
       result = rename(filename, newname);
       if (result!=0) {
          printf("Warning: Could not rename %s\n",filename);
          strncpy(newname, filename, sizeof(newname));
       }
    } else {
       printf("Warning: Could not rename %s\n",filename);
       strncpy(newname, filename, sizeof(newname));
    }
    result = unlink(newname);
    if (result!=0) return 3;
    return 0;
}


int main (int argc, char *argv[])
{
    char rmfile[200];
    int errors = 0;
    int result;
    int secure = 0;
    int schleife;
    if (argc==1){
        printf("Syntax: %s [-ssv] file1 file2 etc.\n", argv[0]);
        printf("%s does a secure overwrite/rename/delete of the target file(s).\n", argv[0]);
        printf("Option -ss does real secure overwrite (39 times), -v is verbose mode.\n");
        exit(1);
    }
    while (1) {
        result = getopt(argc, argv, "SsVvDd");
        if (result<0) break;
        switch (result) {
	    case 'S' :
            case 's' : secure++;
            	       break;
	    case 'V' :
            case 'v' : verbose++;
            	       break;
            case 'D' :
            case 'd' : break;	      
            default :  exit(1);        
        }
    }
    schleife = optind;
    srand(getpid()+getuid());
    while (schleife < argc) {
        strncpy(rmfile, argv[schleife], sizeof(rmfile));
        if (verbose) printf("Deleting %s ", rmfile);
        fflush(stdout);
        result = (int) smash_it(rmfile, secure);
        switch (result) {
            case 0 : if (verbose) printf(" Done\n");
                     break;
            case 1 : printf("%s : File not found or no write permissions\n", rmfile);
            	     break;
            case 2 : printf("%s : Error while writing to file\n", rmfile);
                     break;
            case 3 : printf("Could not delete %s\n", rmfile);
                     break;
            default: printf("Unknown error\n");
        }
        schleife++;
        if (result) errors++;
    }
    (void) fflush(stdout);
    if (errors) return 1;
       else return 0;
}
